"""user.py contains basic information about the steno3d user"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super

import properties


class User(properties.PropertyClass):
    """Class representing a user instance"""
    _model_api_location = "user"
    email = properties.String('Email')
    name = properties.String('Name')
    url = properties.String('URL')
    affiliation = properties.String('Affiliation')
    location = properties.String('Location')
    username = properties.String('Username')

    devel_key = properties.String('Developer API Key')

    file_size_limit = properties.Int(
        'Inidividual file limit',
        default=5000000
    )
    project_size_limit = properties.Int(
        'Project size limit',
        default=25000000
    )
    project_resource_limit = properties.Int(
        'Maximum resources in a project',
        default=25
    )

    def _on_prop_change(self, prop, pre, post):
        if not pre == post:
            print('User data must be modified in settings at steno3d.com')
            setattr(self, '_p_' + prop, pre)

    def login_with_json(self, login_json):
        setattr(self, '_p_username', login_json['uid'])
        setattr(self, '_p_email', login_json['email'])
        setattr(self, '_p_name', login_json['name'])
        setattr(self, '_p_url', login_json['url'])
        setattr(self, '_p_affiliation', login_json['affiliation'])
        setattr(self, '_p_location', login_json['location'])

    def set_key(self, devel_key):
        setattr(self, '_p_devel_key', devel_key)

    def logout(self):
        setattr(self, '_p_username', None)
        setattr(self, '_p_email', None)
        setattr(self, '_p_name', None)
        setattr(self, '_p_url', None)
        setattr(self, '_p_affiliation', None)
        setattr(self, '_p_location', None)
        setattr(self, '_p_devel_key', None)

    @property
    def logged_in(self):
        return self.username is not None



"""base.py sets up BaseExample class that allows examples to be accessed
as properties at the class level
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from os import makedirs
from os import mkdir
from os.path import exists
from os.path import expanduser
from os.path import isdir
from os.path import realpath
from os.path import sep

from future.utils import with_metaclass
from requests import get
from six import string_types
from zipfile import ZipFile


class exampleproperty(object):
    """wrapper that sets class method as saved property"""

    def __init__(self, func):
        self.func = classmethod(func)
        self.propname = '_' + func.__name__

    def __get__(self, cls, owner):
        if getattr(owner, self.propname, None) is None:
            setattr(owner, self.propname, self.func.__get__(None, owner)())
        return getattr(owner, self.propname)


class exampleoutput(object):
    """wrapper that sets class method to remove saved properties"""

    def __init__(self, func, directory):
        self.func = classmethod(func)
        self._directory = ['_' + prop for prop in directory]
        self._stash = [None for prop in directory]

    def __get__(self, cls, owner):
        def _output():
            """return the original function surrounded by stash and unstash"""
            self.stash(owner)
            out = self.func.__get__(None, owner)()
            self.unstash(owner)
            return out
        return _output

    def stash(self, owner):
        """stash saved properties during example output"""
        for i, prop in enumerate(self._directory):
            self._stash[i] = getattr(owner, prop, None)
            setattr(owner, prop, None)

    def unstash(self, owner):
        """unstash saved properties after example output"""
        for i, prop in enumerate(self._directory):
            setattr(owner, prop, self._stash[i])


class _ExampleMetaClass(type):
    """metaclass that wraps every function with exampleproperty()"""

    def __new__(mcs, name, bases, attrs):
        if name not in ('BaseExample', 'Images'):
            assert ('get_resources' in attrs and
                    callable(attrs['get_resources'])), \
                'Example {cls} class must have method get_resources()'.format(
                    cls=name
                )
            assert ('get_project' in attrs and
                    callable(attrs['get_project'])), \
                'Example class {cls} must have method get_project()'.format(
                    cls=name
                )

        example_directory = []
        for attr in attrs:
            value = attrs[attr]
            if not attr.startswith('get_') and not attr == 'fetch_data' and \
                    callable(value):
                attrs[attr] = exampleproperty(value)
                example_directory.append(attr)

        for attr in attrs:
            value = attrs[attr]
            if attr.startswith('get_') and callable(value):
                attrs[attr] = exampleoutput(value, example_directory)

        return type.__new__(mcs, name, bases, attrs)


class BaseExample(with_metaclass(_ExampleMetaClass, object)):
    """basic class that all examples inherit from"""

    def __init__(self, *args, **kwargs):
        raise TypeError('Examples cannot be instantiated. Please access '
                        'the properties directly.')

    def example_name(self):
        return 'BaseExample'

    def data_directory(self):
        """path to directory containing all assets"""
        return sep.join([expanduser('~'), '.steno3d_client', 'assets'])

    def sub_directory(self):
        return self.example_name.lower()

    def filenames(self):
        return []

    def data_url(self):
        return 'https://storage.googleapis.com/steno3d-examples'

    @classmethod
    def fetch_data(cls, directory=None, download_if_missing=True,
                   filename=None, verbose=True):
        if filename is not None and not isinstance(filename, string_types):
            raise ValueError('filename: must be the name of one file')
        if cls.filenames == []:
            raise ValueError('Example does not require any files')
        if directory is not None and not isinstance(directory, string_types):
            raise ValueError('directory: must be the name of a directory')
        if directory is not None:
            directory = realpath(expanduser(directory))
            if not isdir(directory):
                raise ValueError(
                    '{}: directory does not exist'.format(directory)
                )
            cls._data_directory = directory
        if verbose:
            print('Fetching data...')
        destination = sep.join([cls.data_directory, cls.sub_directory])
        archive = sep.join([destination, cls.sub_directory + '.zip'])
        if download_if_missing and not exists(cls.data_directory):
            makedirs(cls.data_directory)
        if download_if_missing and not exists(destination):
            mkdir(destination)
        if verbose:
            print('Destination: ' + destination)
        filenames = cls.filenames if filename is None else [filename]
        for fname in filenames:
            if verbose:
                print('    Fetching: ' + fname)
            destination_file = sep.join([destination, fname])
            if exists(destination_file):
                if verbose:
                    print('        Local copy found')
                if filename is not None:
                    return destination_file
                continue
            if not exists(archive) and download_if_missing:
                if verbose:
                    print('        Downloading archive...')
                url = '/'.join([cls.data_url, cls.sub_directory + '.zip'])
                resp = get(url, stream=True)
                if resp.status_code != 200:
                    raise IOError('Error downloading {exclass} data: '
                                  '{archfile}'.format(
                                    exclass=cls.example_name,
                                    archfile=archive
                                  ))
                with open(archive, 'wb') as arch:
                    for chunk in resp:
                        arch.write(chunk)
                if verbose:
                    print('        Archive downloaded successfully')
            if exists(archive):
                if verbose:
                    print('        Local archive found: extracting...')
                try:
                    zf = ZipFile(archive)
                    zf.extract(fname, destination)
                    if verbose:
                        print('        File extracted successfully')
                    if filename is not None:
                        return destination_file
                except KeyError:
                    raise IOError(
                        """Required file(s) not found in archive:
                        Archive file may be out of date. Please delete archive
                        and {exclass}.fetch_data() again: {archfile}""".format(
                            exclass=cls.example_name,
                            archfile=archive
                        )
                    )
                continue
            raise IOError(
                """Required file(s) not found:
                Please call {exclass}.fetch_data() to download and
                save to a default folder in your home directory or
                {exclass}.fetch_data(directory=your_local_directory)
                to set an alternative local directory.""".format(
                    exclass=cls.example_name
                )
            )



if __name__ == '__main__':
    import os
    import glob
    import unittest
    test_file_strings = glob.glob('test_*.py')
    module_strings = [str[0:len(str)-3] for str in test_file_strings]
    suites = [unittest.defaultTestLoader.loadTestsFromName(str) for str
              in module_strings]
    test_suite = unittest.TestSuite(suites)

    unittest.TextTestRunner(verbosity=2).run(test_suite)



"""resource.py contains the base resource classes that user-created
resources depend on in steno3d
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from json import dumps

import properties

from .client import Comms
from .client import needs_login
from .client import plot
from .content import UserContent


class BaseResource(UserContent):
    """Base class for all resources that are added to projects and
    uploaded to steno3d
    """
    meta = properties.Object('Meta data in JSON form.', default={})

    def _get_dirty_data(self, force=False):
        datadict = super()._get_dirty_data(force)
        dirty = self._dirty
        if 'opts' in dirty or (force and hasattr(self, 'opts')):
            datadict['meta'] = self.opts._json
            self.opts._mark_clean(recurse=False)
        return datadict

    @properties.validator
    def validate(self):
        """Check if content is built correctly"""
        return True


    def _validate_file_size(self, name):
        if Comms.user.logged_in:
            file_limit = Comms.user.file_size_limit
            if self.nbytes(name) > file_limit:
                raise ValueError(
                    '{name} file size ({file} bytes) exceeds limit: '
                    '{lim} bytes'.format(name=name,
                                         file=self.nbytes(name),
                                         lim=file_limit)
                )
        return True


class CompositeResource(BaseResource):
    """A composite resource that stores references to lower-level objects."""
    project = properties.Pointer(
        'Project',
        ptype='Project',
        required=True,
        repeated=True
    )
    _children = {
        'mesh': None,
        'data': 'data',
        'textures': None,
    }

    def __init__(self, project=None, **kwargs):
        if project is None:
            raise TypeError('Resource must be constructed with its '
                            'containing project(s)')
        super().__init__(**kwargs)
        self.project = project

    @classmethod
    def _url_view_from_uid(cls, uid):
        """Get full url from a uid"""
        url = '{base}{mapi}/{uid}'.format(
            base=Comms.base_url,
            mapi=cls._model_api_location,
            uid=uid)
        return url

    @properties.validator
    def validate(self):
        for proj in self.project:
            if self not in proj.resources:
                raise ValueError('Project/resource pointers misaligned: '
                                 'Ensure that projects contain all the '
                                 'resources that point to them.')

    @needs_login
    def upload(self, sync=False, verbose=True, print_url=True):
        """Upload the resource through its containing project(s)"""
        for proj in self.project:
            proj.upload(sync, verbose, False)
        if print_url:
            print(self._url)
        return self._url



    def _get_dirty_data(self, force=False):
        datadict = super()._get_dirty_data(force)
        dirty = self._dirty_props
        if 'mesh' in dirty or force:
            datadict['mesh'] = dumps({
                'uid': self.mesh._json['longUid']
            })
        if 'data' in dirty or force:
            datadict['data'] = dumps([
                {
                    'location': d.location,
                    'uid': d.data._json['longUid']
                } for d in self.data
            ])
        if 'textures' in dirty or (force and hasattr(self, 'textures')):
            datadict['textures'] = dumps([
                {
                    'uid': t._json['longUid']
                } for t in self.textures
            ])
        return datadict

    def _upload_dirty(self, sync=False, verbose=True, tab_level=''):
        dirty = self._dirty
        if 'mesh' in dirty:
            self.mesh._upload(sync, verbose, tab_level)
        if 'data' in dirty:
            [d.data._upload(sync, verbose, tab_level) for d in self.data]
        if 'textures' in dirty:
            [t._upload(sync, verbose, tab_level) for t in self.textures]

    def _on_property_change(self, name, pre, post):
        if name == 'project':
            if pre is None:
                pre = []
            if post is None:
                post = []
            for proj in post:
                if proj not in pre and self not in proj.resources:
                    proj.resources += [self]
            for proj in pre:
                if proj not in post and self in proj.resources:
                    proj.resources = [r for r in proj.resources
                                      if r is not self]
            if len(set(post)) != len(post):
                post_post = []
                for p in post:
                    if p not in post_post:
                        post_post += [p]
                self.project = post_post
        super()._on_property_change(name, pre, post)

    @property
    def _url(self):
        if getattr(self, '_upload_data', None) is not None:
            return self._url_view_from_uid(self._upload_data['uid'])

    @property
    @needs_login
    def url(self):
        """steno3d.com url of project if uploaded"""
        if getattr(self, '_upload_data', None) is None:
            print('Resource not uploaded: Please upload() '
                  'before accessing the URL.')
        return self._url

    @needs_login
    def plot(self):
        """Display the 3D representation of the content"""
        if getattr(self, '_upload_data', None) is None:
            print('Resource not uploaded: Please upload() '
                  'before plotting.')
            return
        return plot(self._url)


class BaseMesh(BaseResource):
    """Base class for all mesh resources. These are contained within
    each composite resources and define its structure
    """


class BaseData(BaseResource):
    """Base class for all data resources. These can be contained within
    each composite resource and define data corresponding to the mesh
    """
    @properties.classproperty
    @classmethod
    def _model_api_location(cls):
        """api destination for texture resource"""
        if getattr(cls, '__model_api_location', None) is None:
            cls.__model_api_location = 'resource/data/{class_name}'.format(
                class_name=cls._resource_class)
        return cls.__model_api_location


class BaseTexture2D(BaseResource):
    """Base class for all texture resources. These can be contained
    within some composite resources and define data in space that gets
    mapped to the mesh.
    """
    @properties.classproperty
    @classmethod
    def _model_api_location(cls):
        """api destination for texture resource"""
        if getattr(cls, '__model_api_location', None) is None:
            cls.__model_api_location = 'resource/texture2d/{class_name}'.format(
                class_name=cls._resource_class)
        return cls.__model_api_location



"""line.py contains the Line composite resource for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super

from numpy import max as npmax
from numpy import min as npmin
import properties

from .base import BaseMesh
from .base import CompositeResource
from .options import ColorOptions
from .options import Options


class _Mesh1DOptions(Options):
    pass


class _LineOptions(ColorOptions):
    pass


class Mesh1D(BaseMesh):
    """Contains spatial information of a 1D line set"""
    vertices = properties.Array(
        'Mesh vertices',
        shape=('*', 3),
        dtype=float,
        required=True
    )
    segments = properties.Array(
        'Segment endpoint indices',
        shape=('*', 2),
        dtype=int,
        required=True
    )
    opts = properties.Pointer(
        'Options',
        ptype=_Mesh1DOptions
    )

    @property
    def nN(self):
        """ get number of nodes """
        return len(self.vertices)

    @property
    def nC(self):
        """ get number of cells """
        return len(self.segments)

    def nbytes(self, name=None):
        if name in ('segments', 'vertices'):
            return getattr(self, name).astype('f4').nbytes
        elif name is None:
            return self.nbytes('segments') + self.nbytes('vertices')
        raise ValueError('Mesh1D cannot calculate the number of '
                         'bytes of {}'.format(name))

    def _on_property_change(self, name, pre, post):
        try:
            if name in ('segments', 'vertices'):
                self._validate_file_size(name)
        except ValueError as err:
            setattr(self, '_p_' + name, pre)
            raise err
        super()._on_property_change(name, pre, post)


    @properties.validator
    def validate(self):
        """Check if mesh content is built correctly"""
        if npmin(self.segments) < 0:
            raise ValueError('Segments may only have positive integers')
        if npmax(self.segments) >= len(self.vertices):
            raise ValueError('Segments expects more vertices than provided')
        self._validate_file_size('segments')
        self._validate_file_size('vertices')
        return True

    def _get_dirty_files(self, force=False):
        dirty = self._dirty_props
        files = dict()
        if 'vertices' in dirty or force:
            files['vertices'] = \
                self._properties['vertices'].serialize(self.vertices)
        if 'segments' in dirty or force:
            files['segments'] = \
                self._properties['segments'].serialize(self.segments)
        return files


class _LineBinder(properties.PropertyClass):
    """Contains the data on a 1D line set with location information"""
    location = properties.String(
        'Location of the data on mesh',
        required=True,
        choices={
            'CC': ('LINE', 'FACE', 'CELLCENTER', 'EDGE', 'SEGMENT'),
            'N': ('VERTEX', 'NODE', 'ENDPOINT')
        }
    )
    data = properties.Pointer(
        'Data',
        ptype='DataArray',
        required=True
    )


class Line(CompositeResource):
    """Contains all the information about a 1D line set"""
    mesh = properties.Pointer(
        'Mesh',
        ptype=Mesh1D,
        required=True
    )
    data = properties.Pointer(
        'Data',
        ptype=_LineBinder,
        repeated=True
    )
    opts = properties.Pointer(
        'Options',
        ptype=_LineOptions
    )

    def nbytes(self):
        return self.mesh.nbytes() + sum(d.data.nbytes() for d in self.data)

    @properties.validator
    def validate(self):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(self.data):
            assert dat.location in ('N', 'CC')
            valid_length = (
                self.mesh.nC if dat.location == 'CC'
                else self.mesh.nN
            )
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'line.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        super(Line, self).validate()
        return True


__all__ = ['Line', 'Mesh1D']



"""There are parser modules available for Steno3D to read certain file
types. When a parser module is imported, the available parsers appear
inside `steno3d.parsers`.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from os.path import expanduser as _expanduser
from os.path import isfile as _isfile
from os.path import realpath as _realpath

from future.utils import with_metaclass as _with_metaclass
from properties import PropertyClass as _PropertyClass
from properties import String as _String
from six import string_types as _string_types


class _ParserMetaClass(type):
    """Metaclass to ensure Parser classes fit the requried format and
    get added to the steno3d.parsers namespace
    """

    def __new__(mcs, name, bases, attrs):
        if name != 'BaseParser':
            assert 'extensions' in attrs, \
                "You must give the `extensions` that are supported by the " \
                "{name} parser, e.g. ('obj',)".format(name=name)
            exts = attrs['extensions']
            assert isinstance(exts, tuple) and len(exts) > 0, \
                'The `extensions` in the {name} parser must be a tuple of ' \
                'supported extensions'.format(name=name)
            for ext in exts:
                assert isinstance(ext, _string_types), \
                    'Extensions in the {name} parser must be ' \
                    'strings'.format(name=name)
            assert '__init__' not in attrs, \
                "The BaseParser __init__ function takes the file_name and " \
                "any additional keyword input. Please perform other " \
                "initialization tasks for the {name} parser in " \
                "_initialize()".format(name=name)
            assert 'parse' in attrs and callable(attrs['parse']), \
                "Parser class {name} must contain a parse() method"
        new_class = super(_ParserMetaClass, mcs).__new__(
            mcs, name, bases, attrs
        )
        globals()[name] = new_class
        return new_class


class _BaseParserMetaClass(_ParserMetaClass,
                           _PropertyClass.__class__):
    """Augmented metaclass for parsers, inherits from properties
    metaclass
    """


class BaseParser(_with_metaclass(_BaseParserMetaClass,
                                 _PropertyClass)):
    """Base class for Steno3D parser objects"""

    extensions = (None,)
    file_name = _String(
        'The main file to parse',
        required=True
    )

    def __init__(self, file_name, **kwargs):
        super(BaseParser, self).__init__(**kwargs)
        if self.extensions[0] is None:
            raise NotImplementedError(
                'Please use the specific parser corresponding to your '
                'file type or `AllParsers`, not `BaseParser`'
            )
        self.file_name = self._validate_file(file_name)
        self._initialize()

    def _validate_file(self, file_name):
        """function _validate_file

        Input:
            file_name - The file to be validated

        Output:
            validated file_name

        _validate_file verifies the file exists and the extension matches
        the parser extension(s) before proceeding. This hook can be
        overwritten to remove or perform different file checks as long as
        it returns the file_name.
        """
        if not isinstance(file_name, _string_types):
            raise IOError('{}: file_name must be a string'.format(file_name))
        file_name = _realpath(_expanduser(file_name))
        if not _isfile(file_name):
            raise IOError('{}: File not found.'.format(file_name))
        if file_name.split('.')[-1] not in self.extensions:
            raise IOError('{name}: Unsupported extension. Supported '
                          'extensions are {exts}'.format(
                             name=file_name,
                             exts='(' + ', '.join(self.extensions) + ')'
                          ))
        return file_name

    def _initialize(self):
        """function _initialize

        _initialize is a hook that is called during parser __init__. The
        BaseParser sets the file_name and any additional keyword
        arguments, but _initialize can be overwritten to perform any
        additional startup tasks
        """

    def parse(self, project=None, **kwargs):
        """function parse

        Optional input:
            project - Preexisting project to add resources to. If not
                      provided, a new project will be created

        Output:
            tuple of project(s) parsed from file_name
        """
        raise NotImplementedError()

    def export(self, project, **kwargs):
        """function export

        Input:
            project - Steno3D project to be exported

        Output:
            file of the supported extension type
        """
        raise NotImplementedError()


class _AllParserMetaClass(type):
    """Metaclass to ensure Director classes fit the requried format and
    get added to the steno3d.parsers namespace
    """

    def __new__(mcs, name, bases, attrs):
        if name != 'AllParsers':
            assert name.startswith('AllParsers_'), \
                "Parser classes that inherit 'AllParsers' such as {name} " \
                "must have names that start with " \
                "'AllParsers_'".format(name=name)
            assert 'extensions' in attrs, \
                "{name} must contain a dictionary of extensions and " \
                "parsers".format(name=name)
            assert isinstance(attrs['extensions'], dict), \
                "{name} extensions must be a dictionary of extensions and " \
                "supporting parser".format(name=name)
            for ext in attrs['extensions']:
                assert isinstance(ext, _string_types), \
                    'Extensions in {name} must be strings'.format(name=name)
                assert issubclass(type(attrs['extensions'][ext]),
                                  _BaseParserMetaClass), \
                    'Extensions in {name} must direct to a ' \
                    'Parser class'.format(name=name)
        new_class = super(_AllParserMetaClass, mcs).__new__(
            mcs, name, bases, attrs
        )
        globals()[name] = new_class
        return new_class


class _BaseAllParserMetaClass(_AllParserMetaClass,
                              _PropertyClass.__class__):
    """Augmented metaclass for parser Directors, inherits from
    properties metaclass
    """


class AllParsers(_with_metaclass(_BaseAllParserMetaClass,
                                 _PropertyClass)):
    """Base class for Steno3D parser objects that parse all
    available file types
    """

    def __new__(cls, filename, **kwargs):
        if getattr(cls, 'extensions', None) is None:
            cls.extensions = dict()
            parser_keys = [
                k for k in globals()
                if (
                    k != '_BaseParser' and
                    issubclass(type(globals()[k]), _BaseParserMetaClass)
                )
            ]
            for k in parser_keys:
                for ext in globals()[k].extensions:
                    if ext not in cls.extensions:
                        cls.extensions[ext] = globals()[k]
                    elif issubclass(type(cls.extensions[ext]),
                                    _BaseParserMetaClass):
                        cls.extensions[ext] = (cls.extensions[ext].__name__ +
                                               ', ' + globals()[k].__name__)
                    else:
                        cls.extensions[ext] = (cls.extensions[ext] +
                                               ', ' + globals()[k].__name__)

        for ext in cls.extensions:
            if filename.split('.')[-1] == ext:
                if not issubclass(type(cls.extensions[ext]),
                                  _BaseParserMetaClass):
                    raise ValueError(
                        '{ext}: file type supported by more than one parser. '
                        'Please specify one of ({parsers})'.format(
                            ext=ext,
                            parsers=cls.extensions[ext]
                        )
                    )
                return cls.extensions[ext](filename, **kwargs)

        raise ValueError(
            '{bad}: unsupported file extensions. Must be in ({ok})'.format(
                bad=filename.split('.')[-1],
                ok=', '.join(list(cls.extensions))
            )
        )

try:
    del absolute_import, division, print_function, unicode_literals
except NameError:
    # Error cleaning namespace
    pass



"""point.py contains the Point composite resource for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super

import properties

from .base import BaseMesh
from .base import CompositeResource
from .options import ColorOptions
from .options import Options


class _Mesh0DOptions(Options):
    pass


class _PointOptions(ColorOptions):
    pass


class Mesh0D(BaseMesh):
    """Contains spatial information of a 0D point cloud."""
    vertices = properties.Array(
        'Point locations',
        shape=('*', 3),
        dtype=float,
        required=True
    )
    opts = properties.Pointer(
        'Mesh0D Options',
        ptype=_Mesh0DOptions
    )

    @property
    def nN(self):
        """ get number of nodes """
        return len(self.vertices)

    def nbytes(self, name=None):
        if name is None or name == 'vertices':
            return self.vertices.astype('f4').nbytes
        raise ValueError('Mesh0D cannot calculate the number of '
                         'bytes of {}'.format(name))

    def _on_property_change(self, name, pre, post):
        try:
            if name == 'vertices':
                self._validate_file_size(name)
        except ValueError as err:
            setattr(self, '_p_' + name, pre)
            raise err
        super()._on_property_change(name, pre, post)

    @properties.validator
    def validate(self):
        """Check if mesh content is built correctly"""
        self._validate_file_size('vertices')
        return True

    def _get_dirty_files(self, force=False):
        dirty = self._dirty_props
        files = dict()
        if 'vertices' in dirty or force:
            files['vertices'] = \
                self._properties['vertices'].serialize(self.vertices)
        return files


class _PointBinder(properties.PropertyClass):
    """Contains the data on a 0D point cloud"""
    location = properties.String(
        'Location of the data on mesh',
        default='N',
        required=True,
        choices={
            'N': ('NODE', 'CELLCENTER', 'CC', 'VERTEX')
        }
    )
    data = properties.Pointer(
        'Data',
        ptype='DataArray',
        required=True
    )


class Point(CompositeResource):
    """Contains all the information about a 0D point cloud"""
    mesh = properties.Pointer(
        'Mesh',
        ptype=Mesh0D,
        required=True
    )
    data = properties.Pointer(
        'Data',
        ptype=_PointBinder,
        repeated=True
    )
    textures = properties.Pointer(
        'Textures',
        ptype='Texture2DImage',
        repeated=True
    )
    opts = properties.Pointer(
        'Options',
        ptype=_PointOptions
    )

    def nbytes(self):
        return (self.mesh.nbytes() + sum(d.data.nbytes() for d in self.data) +
                sum(t.nbytes() for t in self.textures))

    @properties.validator
    def validate(self):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(self.data):
            assert dat.location == 'N'
            valid_length = self.mesh.nN
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'point.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        super(Point, self).validate()
        return True


__all__ = ['Point', 'Mesh0D']
